import sys

from sage.all import *

import matrix_tools
import utils


class curvature():
    """Calculate curvatures for a metric.
    Usage:
    -- Default: import gr; metric=[[...],...]; gr.base.curvature(metric).
    -- Extensions:
       -- curvature(metric, coord(tau, x1, x2, x3)) # default: t, x, y, z
       -- curvature(metric, simp=lambda x: x.simplify_full())
    """
    def __init__(self, metric, 
                 coord = var('t x y z'),
                 simp=utils.simp
                 ):
        print("Initializing curvature calculation...")
        self.coord = coord
        self.metric = metric
        self.simp = simp

        self.cal_inverse_metric()
        self.cal_sqrtg()
        self.cal_affine()
        self.cal_riemann()
        self.cal_ricci()
        self.cal_R()
        self.curvatures = (self.inverse_metric, 
                           self.sqrtg,
                           self.affine,
                           self.riemann,
                           self.ricci,
                           self.R)

    def cal_inverse_metric(self):
        print("Calculating inverse metric...")
        if isinstance(self.metric, list):
            if len(self.metric) == 4:
                self.inverse_metric = \
                    matrix_tools.dim4sym_invert(self.metric, self.simp)
            else:
                print ("Dimension "+str(len(self.metric))+" not supported yet.")
                sys.exit()
        else: # assume matrix object in sage
            self.inverse_metric = self.metric.inverse().apply_map(self.simp)


    def cal_affine(self):
        print("Calculating affine...")
        # indices ordered as Gamma^{lam}_{mu nu}
        # here mu, ... are integers
        g = self.metric
        G = self.inverse_metric
        x = self.coord # x is upper index
        d = range(len(self.coord))
        # as a list, affine = [[(matrix with lambda = 0)], [(lambda=1)], ...]
        # thus affine[0] returns (matrix with lambda = 0)
        self.affine = [[[self.simp(
                    sum([ G[lam][rho]*(
                                diff(g[nu][rho],x[mu])
                                + diff(g[mu][rho],x[nu])
                                - diff(g[mu][nu],x[rho])
                                ) for rho in d]
                    )/Integer(2)
                    ) for nu in d] for mu in d] for lam in d]


    # Gravity sector
    def cal_riemann(self):
        print("Calculating riemann tensor...")
        x = self.coord
        Ga = self.affine
        d = range(len(self.coord))
        self.riemann = [[[[self.simp(
                            diff(Ga[lam][mu][sig],x[nu])
                            - diff(Ga[lam][mu][nu],x[sig])
                            + sum([
                                Ga[eta][mu][sig] * Ga[lam][eta][nu]
                                - Ga[eta][mu][nu] * Ga[lam][eta][sig]
                                for eta in d])
                        ) for sig in d] for nu in d] for mu in d] for lam in d]


    def cal_ricci(self):
        print("Calculating ricci tensor...")
        d = range(len(self.coord))
        self.ricci = [[ self.simp(
                sum([ self.riemann[lam][mu][lam][nu] for lam in d])
                ) for nu in d] for mu in d]


    def cal_R(self):
        print("Calculating ricci scalar R...")
        d = range(len(self.coord))
        G = self.inverse_metric
        self.R = self.simp(sum([ G[mu][nu] * self.ricci[mu][nu]
                for nu in d for mu in d]))


    def cal_sqrtg(self):
        print("Calculating sqrt(-g)...")
        if isinstance(self.metric, list):
            if len(self.metric) == 4:
                detg = matrix_tools.dim4sym_det(self.metric, self.simp)
                self.sqrtg = self.simp(sqrt(-1*detg))
            else:
                print ("Dimension "+str(len(self.metric))+" not supported yet.")
        else: # assume matrix object in sage
            self.sqrtg = self.simp(sqrt(-1*self.metric.det()))
        self.sqrtg = self.sqrtg.simplify_radical()


    def show_affine(self):
        utils.show_components(self.affine,'Gamma(up,dn,dn) ')


    def show_riemann(self):
        utils.show_components(self.riemann,'R(up,dn,dn,dn) ')


    def show_ricci(self):
        utils.show_components(self.ricci,'R(up,dn) ')


class scalar:
    ''' TBA.'''

    def __init__(self, inverse_metric, affine=None, 
                 field = function('phi', var('t')),
                 coord = var('t x y z'), simp=utils.simp):
        print("Initializing scalar sector calculation...")
        self.coord = coord
        self.inverse_metric = inverse_metric
        self.phi = field
        self.simp = simp
        self.cal_X()
        if isinstance(affine, list):
            self.affine = affine
            self.cal_scalar_box()
        

    def cal_X(self):
        print("Calculating scalar field kinetic term X...")
        d = range(len(self.coord))
        G = self.inverse_metric
        x = self.coord # x is upper index
        self.X = self.simp(sum(
                [- G[mu][nu] * diff(self.phi,x[mu]) * diff(self.phi,x[nu]) /SR(2)
                 for nu in d for mu in d]))

        
    def cal_scalar_box(self):
        print("Calculating Box acting on a scalar field ...")
        d = range(len(self.coord))
        G = self.inverse_metric
        Ga = self.affine
        x = self.coord # x is upper index
        self.scalar_box = self.simp(sum(
                [G[mu][nu] * ( 
                        diff(self.phi,x[mu],x[nu]) 
                        - sum([Ga[rho][mu][nu]*diff(self.phi,x[rho])
                               for rho in d])
                        )
                 for nu in d for mu in d]))

